require "engine.class"
local DamageType = require "engine.DamageType"
local Map = require "engine.Map"
local Target = require "engine.Target"
local Talents = require "engine.interface.ActorTalents"
local Chat = require "engine.Chat"

--- Interface to add ToME combat system
module(..., package.seeall, class.make)

--- Checks what to do with the target
-- Talk ? attack ? displace ?
function _M:bumpInto(target)
    local reaction = self:reactionToward(target)
    if reaction < 0 then
        return self:attackTarget(target)
    elseif reaction >= 0 then
        -- Talk ?
        if self.player and target.can_talk then
            local chat = Chat.new(target.can_talk, target, self)
            chat:invoke()
        elseif target.player and self.can_talk then
            local chat = Chat.new(self.can_talk, self, target)
            chat:invoke()
        -- Displace
        elseif self.move_others then
            game.level.map:remove(self.x, self.y, Map.ACTOR)
            game.level.map:remove(target.x, target.y, Map.ACTOR)
            game.level.map(self.x, self.y, Map.ACTOR, target)
            game.level.map(target.x, target.y, Map.ACTOR, self)
            self.x, self.y, target.x, target.y = target.x, target.y, self.x, self.y
        end
    end
end


--- Makes the death happen!
function _M:attackTarget(target, mult)
	if self.combat then
		if self:checkHit(self.combat_atk, target.combat_def) then
			local dam = self.combat.dam + self:getStr() - target.combat_armor
			self:project(self, target.x, target.y, DamageType.PHYSICAL, dam)
		else
			local srcname = game.level.map.seens(self.x, self.y) and self.name:capitalize() or "Something"
			game.logSeen(self, "%s misses %s.", srcname, target.name)
		end
	end

	-- We use up our own energy
	self:useEnergy(game.energy_to_act)

end

--- Computes a logarithmic chance to hit, opposing chance to hit to chance to miss
function _M:checkHit(atk, def, factor)
	print("checkHit", atk, def)
	if atk == 0 then atk = 1 end
	local hit = nil
	factor = factor or 5
	if atk > def then
		local d = atk - def
		hit = math.log10(1 + 5 * d / 50) * 100 + 50
	else
		local d = def - atk
		hit = -math.log10(1 + 5 * d / 50) * 100 + 50
	end
	hit = util.bound(hit, 5, 95)
	print("=> chance to hit", hit)
	return rng.percent(hit), hit
end

function _M:combatDamage(weapon, target)
	local dam = weapon.combat.dam
	dam = rng.range(dam, dam * weapon.combat.dam_range)

	local crit = rng.percent(weapon.combat.crit_chance)
	local srcname = game.level.map.seens(self.x, self.y) and self.name:capitalize() or "Something"
	if crit then game.logSeen(self, "%s preforms a critical hit!", srcname) end
	if crit then dam = dam * 2 end

	dam = dam + self:getDex() - 10 - target.combat_armor	
	
	return dam
end

function _M:combatDamageStr(weapon, target)
	local weapon = weapon
	local dam = combatDamage(weapon) + self:getStr(10) - target.combat_armor
	return math.max(0,dam)
end

function _M:combatDamageDex(weapon, target)
	local dam = combatDamage(weapon, target) + self:getDex(10) - target.combat_armor
	return math.max(0,dam)
end

-- Ranged Attack functions
function _M:rangedAttack(target, talent, tg)
	local weapon = self:hasRangedWeapon()

	if not weapon then
		game.logPlayer(self, "You need to wield a gun to shoot.")
		return nil
	end

	local tg = tg or {type="bolt"}
	tg.range = tg.range or weapon.combat.range
	tg.radius = weapon.combat.radius
	tg.talent = tg.talent or talent
	tg.weapon = tg.weapon or weapon

	local x, y, target = self:getTarget(tg)
	if not x or not y then return nil end

	self:projectile(tg, x, y, gun_projectile)
	
	self:useEnergy(game.energy_to_act)
end

-- Do not call directly! use rangedAttack()!
function gun_projectile(tx, ty, tg, self)
	local weapon = tg.weapon

	local target = game.level.map(tx, ty, game.level.map.ACTOR)
	if not target then return end

	if self:checkHit(self.combat_atk, target.combat_def) then
		self:project(tg, target.x, target.y, DamageType.PHYSICAL, self:combatDamage(weapon, target), {type="gun"})
	else
		local srcname = game.level.map.seens(self.x, self.y) and self.name:capitalize() or "Something"
		game.logSeen(self, "%s misses %s.", srcname, target.name)
	end

end

function _M:hasRangedWeapon(type)
	if not self:getInven("MAINHAND") then return end
	local weapon = self:getInven("MAINHAND")[1]
	if not weapon or not weapon.ranged then
		return nil
	end
	return weapon
end

-- Grenade functions. may merge into the ranged attack function.
function _M:grenadeAttack(target, talent, tg)
	local ammo = self:hasGrenade()

	if not ammo then
		game.logPlayer(self, "You need to have grenades in your belt.")
		return nil
	end

	local tg = tg or {type="ball"}
	if not tg.range then tg.range = math.ceil(3 + self:getStr()/5) end
	tg.radius = ammo.combat.radius
	tg.talent = tg.talent or talent

	local x,y = self:getTarget(tg)
	if not x or not y then return nil end	

	self:project(tg, x, y, DamageType.PHYSICAL, ammo.combat.dam, {type="gun"})

	ammo = self:removeObject(self:getInven("BELT"), 1)
	self:useEnergy(game.energy_to_act)

end

function _M:hasGrenade()
	if not self:getInven("BELT") then return end
	local ammo = self:getInven("BELT")[1]
	if not ammo or ammo.subtype ~= "grenade" then
		return nil
	end
	return ammo
end